/*
 * The contents of this file are subject to the JOnAS Public License Version
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations
 * under
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999,2001 Bull S.A. All Rights Reserved.
 *
 * Contributor(s):
 *
 * --------------------------------------------------------------------------
 * $Id: JConnection.java,v 1.1 2001/12/21 10:53:48 jonas Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_jms;

import java.util.LinkedList;
import javax.jms.Connection;
import javax.jms.ConnectionMetaData;
import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.Session;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

/**
 * Common methods used in JQueueConnection and JTopicConnection.
 */
public class JConnection implements Connection {

    // The XAQueueConnection or XATopicConnection used in the MOM
    protected Connection xac;

    protected boolean closed;
    protected String user;
    protected boolean globaltx = false;
    protected static TransactionManager tm;
    protected JConnectionFactory jcf;
    protected LinkedList sessionlist = new LinkedList();

    /**
     * Constructor
     */
    public JConnection(JConnectionFactory jcf, String user) {
	this.user = user;
	this.jcf = jcf;
	closed = false;
	if (tm == null) {
	    tm = JmsManagerImpl.getTransactionManager();
	}
	// Remember if we are inside a global transaction
	// This is to handle a case that is not expected in JMS specs.
	try {
	    globaltx = (tm.getTransaction() != null);
	} catch (Exception e) {
	}
    }

    // -----------------------------------------------------------------------
    // internal methods
    // -----------------------------------------------------------------------

    /**
     * A new non transacted session has been opened
     */
    protected synchronized boolean sessionOpen(Session s) {
	if (!closed) {
	    sessionlist.add(s);
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * A non transacted session has beem closed
     */
    protected synchronized void sessionClose(Session s) {
	sessionlist.remove(s);
	if (sessionlist.size() == 0 && closed) {
	    notify();
	}
    }

    /**
     * Return the user associated to this connection
     */
    public String getUser() {
	return user;
    }

    // -----------------------------------------------------------------------
    // Connection implementation
    // -----------------------------------------------------------------------

    /**
     * When this method is invoked it should not return until message processing 
     * has been orderly shut down. This means that all message listeners that may 
     * have been running have returned and that all pending receives have returned. 
     * A close terminates all pending message receives on the connection's sessions' 
     * consumers. 
     * @throws JMSException - if JMS implementation fails to return the client ID for this 
     * Connection due to some internal
     */
    public void close() throws JMSException {
	TraceJms.debug( "JConnection : close");
	closed = true;
	if (globaltx) {
	    // Connection that was open inside a global transaction.
	    // Don't wait (to avoid deadlocks) and don't close it now
	    // Since this situation is not expected by the specs, we just
	    // pool this connection for now, waiting better...
	    jcf.freeJConnection(this);
	} else {
	    // Wait for all NON transacted sessions to be finished.
	    // LATER: Should rollback first all transacted sessions still running.
	    synchronized(this) {
		while (sessionlist.size() > 0) {
		    try {
			wait();
		    } catch (InterruptedException e) {
			TraceJms.error("JConnection: Interrupted");
		    }
		}
	    }
	    xac.close();
	}
    }

    /**
     * Get the client identifier for this connection. This value is JMS Provider specific. 
     * Either pre-configured by an administrator in a ConnectionFactory or assigned dynamically 
     * by the application by calling setClientID method.
     * @return the unique client identifier.
     * @throws JMSException - if JMS implementation fails to return the client ID for this 
     * Connection due to some internal
     */
    public String getClientID() throws JMSException {
	TraceJms.debug( "JConnection: getClientID");
	return xac.getClientID();
    }

    /**
     * Set the client identifier for this connection. 
     * If another connection with clientID is already running when this method is called, 
     * the JMS Provider should detect the duplicate id and throw InvalidClientIDException.
     * @param clientID - the unique client identifier
     * @throws JMSException - general exception if JMS implementation fails to set the client 
     * ID for this Connection due to some internal error.
     * @throws InvalidClientIDException - if JMS client specifies an invalid or duplicate client id.
     * @throws IllegalStateException - if attempting to set a connection's client identifier at 
     * the wrong time or when it has been administratively configured.
     */
    public void setClientID(String clientID) throws JMSException {
	TraceJms.debug( "JConnection: setClientID");
	xac.setClientID(clientID);
    }

    /**
     * Get the meta data for this connection.
     * @return the connection meta data.
     * @throws JMSException - general exception if JMS implementation fails to get the Connection 
     * meta-data for this Connection.
     */
    public ConnectionMetaData getMetaData() throws JMSException {
	TraceJms.debug( "JConnection: getMetaData");
	return xac.getMetaData();
    }

    /**
     * Get the ExceptionListener for this Connection.
     * @return the ExceptionListener for this Connection.
     * @throws JMSException - general exception if JMS implementation fails to get 
     * the Exception listener for this Connection.
     */
    public ExceptionListener getExceptionListener() throws JMSException {
	TraceJms.debug( "JConnection: getExceptionListener");
	return xac.getExceptionListener();
    }

    /**
     * Set an exception listener for this connection. 
     * @param listener - the exception listener.
     * @throws JMSException - general exception if JMS implementation fails to set 
     * the Exception listener for this Connection.
     */
    public void setExceptionListener(ExceptionListener listener) throws	JMSException {
	TraceJms.debug( "JConnection  setExceptionListener");
	xac.setExceptionListener(listener);
    }

    /**
     * Start (or restart) a Connection's delivery of incoming messages. 
     * @throws JMSException - if JMS implementation fails to start the message 
     * delivery due to some internal error.
     */
    public void start() throws JMSException {
	TraceJms.debug( "JConnection: start");
	xac.start();
    }

    /**
     * Used to temporarily stop a Connection's delivery of incoming messages. 
     * It can be restarted using its start method.
     * When stopped, delivery to all the Connection's message consumers is inhibited: 
     * synchronous receive's block and messages are not delivered to message listeners. 
     * @throws JMSException - if JMS implementation fails to start the message 
     * delivery due to some internal error.
     */
    public void stop() throws JMSException {
	TraceJms.debug( "JConnection: stop");
	xac.stop();
    }
}
